"""
File: extension.py
------------------
This program recreates the popular Wordle game in the console,
where players attempt to guess a secret WORD_LENGTH letter word
within a given number of tries, with hints given along the way
in the form of colored emoji squares. The program works for any
set WORD_LENGTH between 1 and 10, and any number of MAX_ATTEMPTS.
"""


import random


WORD_LENGTH = 8         # Length of word to be guessed, works from length 1 to 10
MAX_ATTEMPTS = 9        # Number of attempts before the user loses


def main():
    """
    This program creates the Wordle game for the user, where the user has
    MAX_ATTEMPTS attempts to try to guess a random word of WORD_LENGTH. With
    every guess, the user is given hints about which letters are correct or
    incorrect, in addition to feedback about which letters have yet to be used.
    """
    guess_board = create_guess_board()
    answer = choose_answer()
    # answer is a randomly chosen word of length MAX_LENGTH.
    won = False
    attempts = 0
    print_wordle_intro()
    while attempts < MAX_ATTEMPTS:
        if attempts != (MAX_ATTEMPTS - 1):
            guess = input("You have " + str((MAX_ATTEMPTS - attempts)) + " attempts left. Enter your guess: ")
            # General case when the user has more than one attempt left.
        else:
            guess = input("You have 1 attempt left - last try! Enter your guess: ")
            # Special case when the user is on the last attempt to guess the Wordle.
        attempts += 1
        while guess.isalpha() == False or len(guess) != WORD_LENGTH:
            # Displays error message and asks for new input if user input includes anything other than
            # letters or is longer/shorter than the word length.
            print("Invalid input. Your guess should be a " + str(WORD_LENGTH) + " letter word only.")
            guess = input("Please try again: ")
        guess_board = update_guess_board(guess_board, answer, guess, attempts)
        print_guess_status(guess_board)
        # Updates the board with the user's input and prints the guess board with hints to the user.
        if guess.upper() == answer:
            # If the user guesses the word correctly, break out of the loop and display congratulations.
            won = True
            break
    if won == True:
        if attempts == 1:
            print("Congratulations! You got the Wordle in one try - impressive!")
            # Special case when the user guesses the Wordle correctly in one attempt.
        else:
            print("Congratulations! You got the Wordle in " + str(attempts) + " tries!")
            # General case when the user guesses the Wordle correctly in 2+ attempts.
    else:
        print("Sorry, you didn't get the Wordle. The correct answer was: " + str(answer) + ".")
        print("Try again!")
        # Prints a "you lose" message when the user doesn't guess the Wordle after MAX_ATTEMPTS tries.


def print_wordle_intro():
    """
    Prints an introduction to the Wordle game and how it works.
    """
    print("Welcome to Wordle!")
    print("Try to guess the " + str(WORD_LENGTH) + " letter word in " + str(MAX_ATTEMPTS) + " tries.")
    print("Each guess must be a valid " + str(WORD_LENGTH) + " letter word.")
    print("After each guess, you will receive feedback to show how close your guess was to the word.")
    print("🟫 next to your letter means that letter is not in the word.")
    print("🟨 next to your letter means that letter is in the word, but in the wrong spot.")
    print("🟩 next to your letter means that letter is correct.")
    print("Happy guessing!")


def create_guess_board():
    """
    Creates the grid (list of lists) for the guess board with width
    WORD_LENGTH and height MAX_ATTEMPTS, and populates the board with
    "blank" squares (🟫).
    """
    grid = []
    for y in range(MAX_ATTEMPTS):
        guess = []
        for x in range(WORD_LENGTH):
            guess.append(" 🟫 ")
        grid.append(guess)
    return grid


def update_guess_board(guess_board, answer, guess, attempts):
    """
    Given a guess from the user, updates the board with the user's
    guess in a new row, in addition to the corresponding color hints
    for each letter. This only modifies the board, and does not
    display anything to the user.
    """
    row = attempts - 1      # - 1 to match index starting from 0
    spaced_answer = space_out(answer.upper())
    split_answer = spaced_answer.split()
    spaced_guess = space_out(guess.upper())
    split_guess = spaced_guess.split()
    # ^ Converts the answer and guess strings into lists, with the letters of the strings as elements.
    for x in range(WORD_LENGTH):
        # Goes through a loop for every letter in the user's guess.
        if split_guess[x] in split_answer:
            if split_guess[x] == split_answer[x]:
                guess_board[row][x] = split_guess[x] + " 🟩"
                # Adds a green block to the letter if it is in the answer and in the correct spot.
            else:
                guess_board[row][x] = split_guess[x] + " 🟨"
                # Adds a yellow block to the letter if it is in the answer but in the wrong spot.
        else:
            guess_board[row][x] = split_guess[x] + " 🟫"
            # Adds a brown block to the letter if it is not in the answer.
    remove_extra_yellow(guess_board, row, split_answer)
    print_guess_board(guess_board)
    print("")
    return guess_board


def remove_extra_yellow(guess_board, row, split_answer):
    """
    Removes extra yellow letters to avoid displaying extra yellow blocks
    when a letter in the user's guess is in the wrong place in the answer
    multiple times.

    Ex. For answer "ALOHA" and a guess "AAAAA", the desired hint would be
    🟩🟫🟫🟫🟩 rather than 🟩🟨🟨🟨🟩 to show that there are only two "A"s
    in the correct answer (without this function, update_guess_board() gives
    the latter hint, which can be confusing to the users. ).
    """
    for x in range(WORD_LENGTH):
        letter = guess_board[row][x][0]
        ans_freq = get_freq(split_answer, letter)
        # For every letter in the user's guess, see how many time that letter appears in the answer.
        for i in range(WORD_LENGTH):
            if guess_board[row][i] == letter + " 🟩":
                ans_freq -= 1
                # For every correct guess/occurrence of that letter, subtract the ans_freq by 1.
        for j in range(WORD_LENGTH):
            if guess_board[row][j] == letter + " 🟨":
                if ans_freq > 0:
                    ans_freq -= 1
                    # While the ans_freq > 0, subtract the ans_freq by 1 for any yellow of that letter
                    # while leaving the yellow.
                else:
                    guess_board[row][j] = letter + " 🟫"
                    # Once the ans_freq <= 0, there are more yellows than actual occurrences of the
                    # letter, and those additional yellows are turned into browns.


def get_freq(split_list, guess_letter):
    """
    Given a list of strings (in this case, letters in the correct word) and
    a letter, returns the number of times that letter appears in the list.

    >>> split_list = ["K", "N", "O", "L", "L"]
    >>> guess_letter = "L"
    >>> get_freq(split_list, guess_letter)
    2
    >>> guess_letter = "N"
    >>> get_freq(split_list, guess_letter)
    1
    """
    freq = 0
    for letter in split_list:
        if letter == guess_letter:
            freq += 1
    return freq


def print_guess_board(guess_board):
    """
    Prints the guess_board in terminal in an easy-to-read array, with the
    color hints for each guess.
    """
    rows = len(guess_board)
    cols = len(guess_board[0])
    for y in range(rows):
        for x in range(cols):
            letter = guess_board[y][x]
            if x == 0:
                print(" | ", end="")
            if x < (WORD_LENGTH - 1):
                print(str(letter) + " | ", end="")
            if x == (WORD_LENGTH - 1):
                print(str(letter) + " | ")
    return guess_board


def print_guess_status(guess_board):
    """
    Updates and displays a list of letters to the user that are green
    (in word, correct spot), yellow (in word, wrong spot), red "x" or
    cross emoji (guessed by user, not in word), and brown (not yet
    guessed by user, may or may not be in word).
    """
    rows = len(guess_board)
    cols = len(guess_board[0])
    brown_list = ["A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M",
                  "N", "O", "P", "Q", "R", "S", "T", "U", "V", "W", "X", "Y", "Z"]
    # The brown_list starts with all letters as a default for when no guesses have yet been made.
    cross_list = []
    yellow_list = []
    green_list = []
    for y in range(rows):
        for x in range(cols):
            # Loops through every string in the grid (including emojis)
            letter_status = guess_board[y][x]       # Refers to the letter + emoji with a space between
            letter = letter_status[0]               # Refers to only the letter
            letter_emoji = letter_status[2]         # Refers to only the emoji
            if letter != " ":
                if letter_emoji == "🟨" and letter not in yellow_list:
                    yellow_list.append(letter)
                    while letter in brown_list:
                        brown_list.remove(letter)
                    # Moves letters marked with yellow emojis from the brown list to the yellow list.
                elif letter_emoji == "🟩" and letter not in green_list:
                    green_list.append(letter)
                    while letter in brown_list:
                        brown_list.remove(letter)
                    # Moves letters marked with green emojis from the brown list to the green list.
                else:
                    cross_list.append(letter)
                    while letter in brown_list:
                        brown_list.remove(letter)
                    # Moves letters that have been guessed but marked brown from the brown list to
                    # the red cross list.
            for letter in green_list:
                while letter in yellow_list:
                    yellow_list.remove(letter)
                while letter in cross_list:
                    cross_list.remove(letter)
                # Removes any duplicates from the yellow/red cross with the green list.
            for letter in yellow_list:
                while letter in cross_list:
                    cross_list.remove(letter)
                # Removes any duplicates from the red cross list with the yellow list.
    sort_and_print_list(green_list, "🟩:")
    sort_and_print_list(yellow_list, "🟨:")
    sort_and_print_list(cross_list, "❌ :")
    sort_and_print_list(brown_list, "🟫:")
    print("")


def sort_and_print_list(emoji_list, emoji):
    """
    Given a list of letters and an emoji, sorts the list of letters
    in alphabetical order, removes duplicates, and prints the letters.
    """
    emoji_list.sort()
    new_emoji_list = []
    for letter in emoji_list:
        if letter not in new_emoji_list:
            new_emoji_list.append(letter)
        # Removes duplicates in emoji_list by making a new list new_emoji_list.
    emoji_string = ""
    for letter in new_emoji_list:
        emoji_string += " " + letter
        # Turns the list into a string displaying letters alphabetically with a space in between.
    print(emoji + emoji_string)


def space_out(word):
    """
    Given a word, returns a string with the same word with spaces
    between each letter.

    >>> space_out("WATERMELON")
    'W A T E R M E L O N'
    """
    spaced_word = ""
    for ch in word:
        spaced_word = spaced_word + ch + " "
    spaced_word = spaced_word.strip()
    return spaced_word


def choose_answer():
    """
    Based on the value of WORD_LENGTH as set as a constant, chooses a word
    from the proper file with the correct word length.
    """
    if WORD_LENGTH == 1:
        word = choose_from_file("one_letter.txt")
    elif WORD_LENGTH == 2:
        word = choose_from_file("two_letters.txt")
    elif WORD_LENGTH == 3:
        word = choose_from_file("three_letters.txt")
    elif WORD_LENGTH == 4:
        word = choose_from_file("four_letters.txt")
    elif WORD_LENGTH == 5:
        word = choose_from_file("five_letters.txt")
    elif WORD_LENGTH == 6:
        word = choose_from_file("six_letters.txt")
    elif WORD_LENGTH == 7:
        word = choose_from_file("seven_letters.txt")
    elif WORD_LENGTH == 8:
        word = choose_from_file("eight_letters.txt")
    elif WORD_LENGTH == 9:
        word = choose_from_file("nine_letters.txt")
    elif WORD_LENGTH == 10:
        word = choose_from_file("ten_letters.txt")
    return word


def choose_from_file(filename):
    """
    Reads a file full of words that are WORD_LENGTH long, and chooses a
    random word from that file and returns it.
    """
    word_list = []
    with open(filename) as file:
        for line in file:
            line = line.strip()
            word_list.append(line)
    num = random.randint(0, len(word_list) - 1)
    return word_list[num]


if __name__ == '__main__':
    main()
